
/*
 *
 *	s t a t
 *
 *	display file info
 *
 *	usage:   stat [-V] | [-nrwxtugoiUGsamc] | [-f arg] file ...
 *
 *	Goran Larsson
 *
 */

#define VERSION "2.2"
 
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>

#define MAXWIDTH 128
#define FUZZ 16

struct passwd *getpwuid ();
struct group  *getgrgid ();

char *ME;
char outbuffer[BUFSIZ];
char m[MAXWIDTH+FUZZ];

/* from libc.a */

int strlen ();
char *strcpy ();
char *strcat ();
char *strrchr ();

/* from filetools.o */

char *Fpath ();		/* get full path to a file name /*
int Fnorm ();		/* normalize a path (remove excess "/", ".." and ".") */

/* from getopt.c or libc.a */

int getopt ();
int optind;
char *optarg;

/* this file */

int makefull ();
int display ();
char *putprot ();
int putmode ();

usage ()
{
	fprintf (
	  stderr,
	  "usage: %s [-V] [-nNrwxtugoiUGsamc] [-f arg] file..\n",
	  ME
	);
	fflush ( stdout );
	exit ( 1 );
}

main ( argc, argv )
int argc;
char *argv[];

{
	struct stat buf;
	char fyle[256];
	int c, i, j;
	char expanded = 0;
	char do_sep   = 0;
	char do_name  = 0;
	char do_Name  = 0;
	char do_rprot = 0;
	char do_wprot = 0;
	char do_xprot = 0;
	char do_type  = 0;
	char do_uprot = 0;
	char do_gprot = 0;
	char do_oprot = 0;
	char do_ino   = 0;
	char do_uid   = 0;
	char do_gid   = 0;
	char do_size  = 0;
	char do_ta    = 0;
	char do_tm    = 0;
	char do_tc    = 0;
	char do_form  = 0;
	char *format;
	int field;
	char infield  = 0;
	char align;
	char *pbuf;

	if ( ME = strrchr ( argv[0], '/' ) ) {
		ME++;
	} else {
		ME = argv[0];
	}

	setbuf ( stdout, outbuffer );

	while ( ( c = getopt ( argc, argv, "VnNrwxtugoiUGsamcf:" ) ) != EOF ) {
		switch ( c ) {
			case 'V' :
				printf ( "stat V%s\n", VERSION );
				fflush ( stdout );
				exit ( 0 );
			case 'n' :
				do_name  = ++do_sep;
				break;
			case 'N' :
				do_Name  = ++do_sep;
				break;
			case 'r' :
				do_rprot = ++do_sep;
				break;
			case 'w' :
				do_wprot = ++do_sep;
				break;
			case 'x' :
				do_xprot = ++do_sep;
				break;
			case 't' :
				do_type  = ++do_sep;
				break;
			case 'u' :
				do_uprot = ++do_sep;
				break;
			case 'g' :
				do_gprot = ++do_sep;
				break;
			case 'o' :
				do_oprot = ++do_sep;
				break;
			case 'i' :
				do_ino   = ++do_sep;
				break;
			case 'U' :
				do_uid   = ++do_sep;
				break;
			case 'G' :
				do_gid   = ++do_sep;
				break;
			case 's' :
				do_size  = ++do_sep;
				break;
			case 'a' :
				do_ta    = ++do_sep;
				break;
			case 'm' :
				do_tm    = ++do_sep;
				break;
			case 'c' :
				do_tc    = ++do_sep;
				break;
			case 'f' :
				do_form  = 1;
				format   = optarg;
				break;
			case '?' :
				usage ();
		}
	}

	if ( optind >= argc ) {
		usage ();
	}

	if ( do_sep && do_form ) {
		fprintf ( stderr, "%s: incompatible options\n", ME );
		usage ();
	}

	for ( i = optind; i < argc; i++ ) {

		expanded = 0;

		if ( ( i > optind ) && !do_sep && !do_form ) {
			printf ( "----------------------------\n" );
		}

		for ( j = 0; j < strlen ( argv[i] ); j++ ) {
			fyle[j] = argv[i][j];
			fyle[j+1] = '\0';
		}

		/*
		** probe file
		*/

		if ( stat ( fyle, &buf )) {
			fprintf ( stderr, "%s: cannot stat %s\n", ME, fyle );
			continue; /* with next file */
		}

		/*
		** handle all switches except -f
		*/

		if ( do_sep ) {
			for ( j = 1; j <= do_sep; j++ ) {

				if ( do_name == j ) {
					printf ( "%s ", argv[i] );
					continue;
				}

				if ( do_Name == j ) {
					if ( !expanded ) {
						makefull ( fyle );
						expanded++;
					}
					printf ( "%s ", fyle );
					continue;
				}

				if ( do_rprot == j ) {
					pbuf = putprot ( 'r', buf.st_mode );
					printf ( "%s ", pbuf );
					continue;
				}

				if ( do_wprot == j ) {
					pbuf = putprot ( 'w', buf.st_mode );
					printf ( "%s ", pbuf );
					continue;
				}

				if ( do_xprot == j ) {
					pbuf = putprot ( 'x', buf.st_mode );
					printf ( "%s ", pbuf );
					continue;
				}

				if ( do_type == j ) {
					pbuf = putprot ( 't', buf.st_mode );
					printf ( "%s ", pbuf );
					continue;
				}

				if ( do_uprot == j ) {
					pbuf = putprot ( 'u', buf.st_mode );
					printf ( "%s ", pbuf );
					continue;
				}

				if ( do_gprot == j ) {
					pbuf = putprot ( 'g', buf.st_mode );
					printf ( "%s ", pbuf );
					continue;
				}

				if ( do_oprot == j ) {
					pbuf = putprot ( 'o', buf.st_mode );
					printf ( "%s ", pbuf );
					continue;
				}

				if ( do_ino == j ) {
					printf ( "%d ", buf.st_ino );
					continue;
				}

				if ( do_uid == j ) {
					printf ( "%d ", buf.st_uid );
					continue;
				}

				if ( do_gid == j ) {
					printf ( "%d ", buf.st_gid );
					continue;
				}

				if ( do_size == j ) {
					printf ( "%d ", buf.st_size );
					continue;
				}

				if ( do_ta == j ) {
					printf (
						"%.24s ",
					        ctime ( &buf.st_atime )
					);
					continue;
				}

				if ( do_tm == j ) {
					printf (
						"%.24s ",
						ctime ( &buf.st_mtime )
					);
					continue;
				}

				if ( do_tc == j ) {
					printf (
						"%.24s ",
						ctime ( &buf.st_ctime )
					);
					continue;
				}

			}
			printf ( "\n" );
			fflush ( stdout );
			continue;
		}

		/*
		** handle switch -f (printf-like formatting)
		*/

		if ( do_form ) {
		    for ( j = 0; j < strlen ( format ); j++ ) {
		    	if ( infield ) {
		    	    switch ( c = format[j] ) {
				case '0' :
				case '1' :
				case '2' :
				case '3' :
				case '4' :
				case '5' :
				case '6' :
				case '7' :
				case '8' :
				case '9' :
				    field = ( field * 10 ) + ( c - '0' );
				    break;
				case 'l' :
				case 'r' :
				case 'c' :
				    align = c;
				    goto postfield;
				    break;
				default :
				    align = c;
				    goto postfield;
			    }
		    	    continue;
		    	}
			switch ( c = format[j] ) {
			    case '%' :
				field = 0;
postfield:			infield = 0;
				switch ( c = format[++j] ) {
				    case 'n' :
					sprintf ( m, "%s", argv[i] );
					display ( m, field, align );
					break;
				    case 'N' :
					if ( !expanded ) {
					    makefull ( fyle );
					    expanded++;
					}
					sprintf ( m, "%s", fyle );
					display ( m, field, align );
					break;
				    case 'r' :
				    case 'w' :
				    case 'x' :
				    case 't' :
				    case 'u' :
				    case 'g' :
				    case 'o' :
					pbuf = putprot ( c, buf.st_mode );
					display ( pbuf, field, align );
					break;
				    case 'i' :
					sprintf ( m, "%d", buf.st_ino );
					display ( m, field, align );
					break;
				    case 'U' :
					sprintf ( m, "%d", buf.st_uid );
					display ( m, field, align );
					break;
				    case 'G' :
					sprintf ( m, "%d", buf.st_gid );
					display ( m, field, align );
					break;
				    case 's' :
					sprintf ( m, "%d", buf.st_size );
					display ( m, field, align );
					break;
				    case 'a' :
					sprintf (
					    m,
					    "%.24s",
					    ctime ( &buf.st_atime )
					);
					display ( m, field, align );
					break;
				    case 'm' :
					sprintf (
					    m,
					    "%.24s",
					    ctime ( &buf.st_mtime )
					);
					display ( m, field, align );
					break;
				    case 'c' :
					sprintf (
					    m,
					    "%.24s",
					    ctime ( &buf.st_ctime )
					);
					display ( m, field, align );
					break;
				    case '0' :
				    case '1' :
				    case '2' :
				    case '3' :
				    case '4' :
				    case '5' :
				    case '6' :
				    case '7' :
				    case '8' :
				    case '9' :
					field = c - '0';
					infield = 1;
					break;
				    case '%' :
					fputc ( '%', stdout );
					break;
				    default :
					if ( field ) {
					    printf (
					      "%%%d%c%c", field, align, c
					    );
					} else {
					    printf ( "%%%c", c );
					}
					break;
				}
				break;
			    case '\\' :
				switch ( c = format[++j] ) {
				    case 'n' :
					fputc ( '\n', stdout );
					break;
				    case 't' :
					fputc ( '\t', stdout );
					break;
				    case '%' :
					fputc ( '%', stdout );
					break;
				    case '\\' :
					fputc ( '\\', stdout );
					break;
				    default  :
					printf ( "\\%c", c );
					break;
				}
				break;
			    default  :
				fputc ( c, stdout );
				break;
			    }
			}
			fflush ( stdout );
			continue;
		}

		/*
		** handle full output (no switches)
		*/

		if ( !expanded ) {
			makefull ( fyle );
			expanded++;
		}
		printf ( "\nFile: %s\n", fyle );

		printf ( "File mode          :  %o \t", buf.st_mode );
		putmode ( buf.st_mode );
		printf ( "\n" );

		printf ( "Inode number       :  %d\n", buf.st_ino );

		printf ( "Device ID          :  %d \t%d,%d\n",
			buf.st_dev,
			( buf.st_dev >> 8 ) & 0xFF,
			buf.st_dev & 0xFF
		);

		printf ( "Special device ID  :  %d \t%d,%d\n",
			buf.st_rdev,
			( buf.st_rdev >> 8 ) & 0xFF,
			buf.st_rdev & 0xFF
		);

		printf ( "Number of links    :  %d\n", buf.st_nlink );

		if ( getpwuid ( buf.st_uid ) == 0 ) {
			printf ( "User ID            :  %d \t(%s)\n",
				buf.st_uid,
				"******"
			);
		} else {
			printf ( "User ID            :  %d \t(%s)\n",
				buf.st_uid,
				getpwuid ( buf.st_uid ) -> pw_name
			);
		}

		if ( getgrgid ( buf.st_gid ) == 0 ) {
			printf ( "Group ID           :  %d \t(%s)\n",
				buf.st_gid,
				"******"
			);
		} else {
			printf ( "Group ID           :  %d \t(%s)\n",
				buf.st_gid,
				getgrgid ( buf.st_gid ) -> gr_name
			);
		}

		printf ( "Size in bytes      :  %d\n", buf.st_size );

		printf ( "Last access        :  %s",
			ctime ( &buf.st_atime )
		);

		printf ( "Last modification  :  %s",
			ctime ( &buf.st_mtime )
		);

		printf ( "Last status change :  %s",
			ctime ( &buf.st_ctime )
		);

		printf ( "\n" );
		fflush ( stdout );
	}

	exit ( 0 );
}

makefull ( fyle )
char *fyle;

{
	int j, fs;
	char path1[256];
	char path2[256];
	char *cp;

	/*
	** convert file name to a fully qualified filename
	*/

	strcpy ( path2, fyle );
	fs = -1;
	for ( j = 0; j < strlen ( path2 ); j++ ) {
		if ( path2[j] == '/' ) {
			fs = j;
		}
	}
	for ( j = 0; j < strlen ( path2 ); j++ ) {
		if ( j < fs ) {
			path1[j] = path2[j];
			path1[j+1] = '\0';
		}
		if ( j > fs ) {
			fyle[j-fs-1] = path2[j];
			fyle[j-fs] = '\0';
		}
	}

	if ( path1[0] == '\0' ) {
		cp = Fpath ( "." );
		strcpy ( path2, cp );
	} else {
		if ( path1[0] != '/' ) {
			cp = Fpath ( path1 );
			strcpy ( path2, cp );
		} else {
			strcpy ( path2, path1 );
		}
	}

	strcat ( path2, "/" );
	strcat ( path2, fyle );
	strcpy ( fyle, path2 );
	Fnorm ( fyle );		/* normalize */
}

display ( buf, width, how )
char *buf;
int width;
char how;

{
	int i, j, sw;
	int trunc;

	trunc = 0;
	sw = strlen ( buf );
	if ( width > MAXWIDTH ) {		/* field to long ?         */
		printf ( "*huh?*" );
		return;
	}
	if ( width == 0 ) {			/* default justification ? */
		width = sw;
		how = 'l';
	}
	if ( sw > width ) {			/* item to long ?          */
		trunc = sw - width;
		sw = width;
	}
	switch ( how ) {
		case 'l' 		:		/* left justify    */
			for ( i = 0; i < sw; i++ ) {
				if ( ( i == ( sw-1 ) ) && trunc ) {
					buf[i] = '!';
				}
				fputc ( buf[i], stdout );
			}
			i = width - sw;
			while ( i-- > 0 ) {
				fputc ( ' ', stdout );
			}
			break;
		case 'r' :				/* right justify   */
			i = width - sw;
			while ( i-- > 0 ) {
				fputc ( ' ', stdout );
			}
			for ( i = 0; i < sw; i++ ) {
				if ( ( i == 0 ) && trunc ) {
					buf[trunc+i] = '!';
				}
				fputc ( buf[trunc+i], stdout );
			}
			break;
		case 'c' :				/* center          */
			i = ( width - sw ) / 2;
			j = width - i - sw;
			if ( trunc ) {
				trunc = trunc / 2;
				if ( trunc ) {
					buf[trunc] = '!';
				}
				buf[trunc+width-1] = '!';
			}
			while ( i-- > 0 ) {
				fputc ( ' ', stdout );
			}
			for ( i = 0; i < sw; i++ ) {
				fputc ( buf[trunc+i], stdout );
			}
			while ( j-- > 0 ) {
				fputc ( ' ', stdout );
			}
			break;
		default :
			printf ( "*jerk*" );
			break;
	}
}

char *putprot ( type, mode )
char type;
int mode;

{
	int i, j, shif, mask;
	static char tmp[4];

	j = mask = 0;
	switch ( type ) {
		case 't' :
			switch ( ( mode >> 12 ) & 0xF ) {
				case 010 : tmp[0] = '-'; break;	/* FREG */
				case 004 : tmp[0] = 'd'; break;	/* FDIR */
				case 002 : tmp[0] = 'c'; break;	/* FCHR */
				case 006 : tmp[0] = 'b'; break;	/* FBLK */
				case 007 : tmp[0] = '7'; break;	/* FMPB */
				case 003 : tmp[0] = '3'; break;	/* FMPC */
				case 005 : tmp[0] = '5'; break;	/* FNAM */
				case 001 : tmp[0] = 'p'; break;	/* FIFO */
				default:   tmp[0] = '?'; break;
			}
			tmp[1] = '\0';
			return ( &tmp[0] );
		case 'r' : mask = 4; break;
		case 'w' : mask = 2; break;
		case 'x' : mask = 1; break;
	}
	if ( mask ) {
		for ( i = 2; i >= 0; i-- ) {
			if ( ( ( mode >> ( i * 3 ) ) & 0x7 ) & mask ) {
				switch ( i ) {
					case 2: tmp[j++] = 'u'; break;
					case 1: tmp[j++] = 'g'; break;
					case 0: tmp[j++] = 'o'; break;
				}
			} else {
				tmp[j++] = ' ';
			}
		}
		tmp[j] = '\0';
		return ( &tmp[0] );
	}
	switch ( type ) {
		case 'u' : shif = 2; break;
		case 'g' : shif = 1; break;
		case 'o' : shif = 0; break;
	}
	for ( i = 2; i >= 0; i-- ) {
		if ( ( mode >> ( ( shif * 3 ) + i ) ) & 1 ) {
			switch ( i ) {
				case 2: tmp[j++] = 'r'; break;
				case 1: tmp[j++] = 'w'; break;
				case 0: tmp[j++] = 'x'; break;
			}
		} else {
			tmp[j++] = '-';
		}
	}
	tmp[j] = '\0';
	return ( &tmp[0] );
}

putmode ( mode )
int mode;
{
	int i;

	for ( i = 2; i >= 0; i-- ) {
		switch ( i ) {
			case 2: printf ( "u:" ); break;
			case 1: printf ( "g:" ); break;
			case 0: printf ( "o:" ); break;
		}
		switch ( ( mode >> ( i * 3 ) ) & 0x7 ) {
			case 000 : printf ( "--- " ); break;
			case 001 : printf ( "--x " ); break;
			case 002 : printf ( "-w- " ); break;
			case 003 : printf ( "-wx " ); break;
			case 004 : printf ( "r-- " ); break;
			case 005 : printf ( "r-x " ); break;
			case 006 : printf ( "rw- " ); break;
			case 007 : printf ( "rwx " ); break;
		}
	}

	switch ( ( mode >> 12 ) & 0xF ) {
		case 010 : printf ( "file   " ); break;		/* FREG */
		case 004 : printf ( "dir    " ); break;		/* FDIR */
		case 002 : printf ( "char   " ); break;		/* FCHR */
		case 006 : printf ( "block  " ); break;		/* FBLK */
		case 007 : printf ( "*FMPB* " ); break;		/* FMPB */
		case 003 : printf ( "*FMPC* " ); break;		/* FMPC */
		case 005 : printf ( "*FNAM* " ); break;		/* FNAM */
		case 001 : printf ( "pipe   " ); break;		/* FIFO */
		default:   printf ( "****** " ); break;
	}

	switch ( ( mode >> 10 ) & 0x3 ) {
		case 000 : printf ( "        " ); break;
		case 001 : printf ( "GID     " ); break;	/* SGID */
		case 002 : printf ( "UID     " ); break;	/* SUID */
		case 003 : printf ( "GID UID " ); break;
	}

	switch ( ( mode >> 9 ) & 0x1 ) {
		case 000 : printf ( "       " ); break;
		case 001 : printf ( "*SVTX* " ); break;		/* SVTX */
	}
}
